#include "level.h"

int abs(int x) {
	if (x < 0) return -x;
	else return x;
}

int checkErrors(const int err) {
	if (err < 0) {
		int nerr = err * -1;

		Bdisp_AllClr_DDVRAM();
		for (int i = 0; i < nerr; ++i) {
			Bdisp_SetPoint_VRAM(i*2 + 1, 1, (unsigned char)1);
		}

		PrintXY(32, 28, (unsigned char*)"ERROR !", 1);
		Bdisp_PutDisp_DD();

		while (!KeyDown(KEY_CTRL_AC)) {
			Sleep(100);
		}

		unsigned int lol = 0;
		GetKey(&lol);

		return err;
	}

	return 0;
}

Level loadLevel(const FONTCHARACTER* path) {
	int handle = Bfile_OpenFile(path, _OPENMODE_READ);
	checkErrors(handle);

	int* p = calloc(4, sizeof(int));
	int err = Bfile_ReadFile(handle, p, sizeof(int) * 4, 0);
	checkErrors(err);

	Level l = {
		.sizeX = p[0],
		.sizeY = p[1],
		.sizeEntities = p[2],
		.sizeMetadata = p[3],
		.level = NULL,
		.entities = NULL,
		.metadata = NULL,
	};

	free(p);

	char* lvl = calloc(l.sizeX * l.sizeY, sizeof(char));
	Entity* ent = calloc(l.sizeEntities, sizeof(Entity));
	Metadata* meta = calloc(l.sizeMetadata, sizeof(Metadata));
	
	err = Bfile_ReadFile(handle, lvl, l.sizeX * l.sizeY, sizeof(int) * 4);
	checkErrors(err);
	l.level = lvl;

	err = Bfile_ReadFile(handle, ent, l.sizeEntities, l.sizeX * l.sizeY + sizeof(int) * 4);
	checkErrors(err);
	l.entities = ent;

	err = Bfile_ReadFile(handle, meta, l.sizeMetadata, l.sizeEntities + l.sizeX * l.sizeY + sizeof(int) * 4);
	checkErrors(err);
	Bfile_CloseFile(handle);

	return l;
}

int saveLevel(Level l, const FONTCHARACTER* path) {
	int err = Bfile_DeleteFile(path);
	checkErrors(err);

	err = Bfile_CreateFile(path, l.sizeX * l.sizeY + l.sizeEntities + l.sizeMetadata + 1);
	checkErrors(err);

	int handle = Bfile_OpenFile(path, _OPENMODE_WRITE);
	int sizes[] = {l.sizeX, l.sizeY, l.sizeEntities, l.sizeMetadata};
	Bfile_WriteFile(handle, sizes, sizeof(sizes));
	Bfile_WriteFile(handle, l.level, l.sizeX * l.sizeY);
	Bfile_WriteFile(handle, l.entities, l.sizeEntities);
	Bfile_WriteFile(handle, l.metadata, l.sizeMetadata);

	Bfile_CloseFile(handle);

	return err;
}

Level newLevel(int sizeX, int sizeY, int sizeEntities, int sizeMetadata) {
	Level l = {
		.sizeX = sizeX,
		.sizeY = sizeY,
		.sizeEntities = sizeEntities,
		.sizeMetadata = sizeMetadata,
		.level = NULL,
		.entities = NULL,
		.metadata = NULL,
	};

	l.level = calloc(sizeX * sizeY, sizeof(char));
	l.entities = calloc(sizeEntities, sizeof(Entity));
	l.metadata = calloc(sizeMetadata, sizeof(Metadata));

	return l;
}

void freeLevel(Level* l) {
	free(l->level);
	free(l->entities);
	free(l->metadata);

	l->level = NULL;
	l->entities = NULL;
	l->metadata = NULL;
}

void editLevel(Level l) {
	int curX = 0; // cursor pos, in world tile
	int curY = 7;
	int posX = 0; // camera pos, in world tile 
	int posY = l.sizeY - 8;

	char currentSprite = 0;
	int lastframetime = RTC_GetTicks();

	unsigned int key = 0; // for the Getkey. Seriously Casio, fuck you
	while (1) {
		if (KeyDown(KEY_CTRL_MENU)) {
			GetKey(&key);
		}

		if (KeyDown(KEY_CTRL_ALPHA)) {
			playLevel(l, (curX + posX) * 256 * 8, (curY + posY) * 256 * 8);
		}

		if (KeyDown(KEY_CTRL_RIGHT)) {
			++curX;
		} else if (KeyDown(KEY_CTRL_LEFT)) {
			--curX;
		}
		if (KeyDown(KEY_CTRL_DOWN)) {
			++curY;								// +y is down !
		} else if (KeyDown(KEY_CTRL_UP)) {
			--curY;
		}

		if (curX >= 16) {
			curX = 15;
			++posX;
		} else if (curX < 0) {
			curX = 0;
			--posX;
		}
		if (curY >= 8) {
			curY = 7;
			++posY;
		} else if (curY < 0) {
			curY = 0;
			--posY;
		}

		if (posX < 0) posX = 0;
		if (posY < 0) posY = 0;
		if (posX > l.sizeX - 16) posX = l.sizeX - 16;
		if (posY > l.sizeY - 8) posY = l.sizeY - 8;

		if (KeyDown(KEY_CTRL_EXE)) {
			l.level[(posX + curX) * l.sizeY + posY + curY] = currentSprite;
		} else if (KeyDown(KEY_CHAR_PMINUS)) {
			l.level[(posX + curX) * l.sizeY + posY + curY] = SPRITE_NONE;
		}

		if (KeyDown(KEY_CHAR_PLUS)) {
			++currentSprite;
		} else if (KeyDown(KEY_CHAR_MINUS)) {
			--currentSprite;
		}

		if (KeyDown(KEY_CTRL_F1)) {
			saveLevel(l, defaultPath);
		} else if (KeyDown(KEY_CTRL_F6)) {
			freeLevel(&l);
			l = loadLevel(defaultPath);
		}

		// draws all the background
		drawScreen(&l, posX, posY, 0, 0);
		// draws currently selected sprite as an indication
		drawSprite(spritedata + currentSprite * 8, 0, 0);
		Bdisp_AreaReverseVRAM(curX * 8, curY * 8, curX * 8 + 7, curY * 8 + 7);
		Bdisp_PutDisp_DD();

		while(lastframetime + 128/8 /* 1sec = 128ticks -> X fps = 128/X */ > RTC_GetTicks()) {
			Sleep(5);
		}

		lastframetime = RTC_GetTicks();
	}
}

void playLevel(Level l, int startPosX, int startPosY) {

	Mario m = {
		.posX = startPosX,	//(4 * 8) * 256,
		.posY = startPosY,	//(l.sizeY * 8 - 4 * 8) * 256, // 4 blocks above the bottom of the stage
		.velX = 0,
		.velY = 0,
		.spr = SPRITE_MARIO_DEFAULT_RIGHT,
		.grounded = 0,
		.bonked = 0,
		.jumping = 0,
		.running = 0,
	};

	int lastposX = m.posX;
	int lastposY = m.posY;
	char inblockUL = 0;
	char inblockUR = 0;
	char inblockDL = 0;
	char inblockDR = 0;
	int jumpFrameCount = 0;
	char wasHoldingJumpButton = 0;
	int isPressingJump = 0;
	int isPressingRun = 0;
	int currentJumpLength = 0;

	int camX = 0; // in pixel coords
	int camY = 0;
	int lastframetime = RTC_GetTicks();
	int framesCount = 0;

	unsigned int key = 0; // for the Getkey. Seriously Casio, fuck you

	while (1) {
		if (KeyDown(KEY_CTRL_MENU)) {
			GetKey(&key);
		}

		if (KeyDown(KEY_CTRL_SHIFT)) return;
 
		if (m.posY > l.sizeY * 256 * 8) return;

		// MARIO START ---------------------------------------------------------------------------------------------------------------------------------------------
		isPressingRun = KeyDown(KEY_CHAR_0);
		// TODO remove this bs
		m.running = isPressingRun;

		if (KeyDown(KEY_CTRL_RIGHT)) {
			//m.posX += 2;
			// TODO: skidding
			m.spr = SPRITE_MARIO_DEFAULT_RIGHT;
			if (m.running) {
				m.velX += MARIO_ACCEL_RUN;
			} else {
				m.velX += MARIO_ACCEL_WALK;
			}
		} else if (KeyDown(KEY_CTRL_LEFT)) {
			//m.posX -= 2;
			m.spr = SPRITE_MARIO_DEFAULT_LEFT;
			if (m.running) {
				m.velX -= MARIO_ACCEL_RUN;
			} else {
				m.velX -= MARIO_ACCEL_WALK;
			}
		}
		/*
		 * if (KeyDown(KEY_CTRL_UP)) {
		 *	m.posY -= 256;
		 * } else if (KeyDown(KEY_CTRL_DOWN)) {
		 *	m.posY += 256;
		 * }
		 */

		// X
		if (m.running) {
			if (m.velX > 0 && m.velX > MARIO_SPEED_RUN_MAX) m.velX = MARIO_SPEED_RUN_MAX;
			if (m.velX < 0 && m.velX < -MARIO_SPEED_RUN_MAX) m.velX = -MARIO_SPEED_RUN_MAX;
		} else {
			if (m.velX > 0 && m.velX > MARIO_SPEED_WALK_MAX) m.velX = MARIO_SPEED_WALK_MAX;
			if (m.velX < 0 && m.velX < -MARIO_SPEED_WALK_MAX) m.velX = -MARIO_SPEED_WALK_MAX;
		}
		m.posX += m.velX;

		if (m.velX > 0) {
			m.velX -= MARIO_ACCEL_SLOWDOWN;
			if (m.velX < 0) m.velX = 0;
		} else if (m.velX < 0) {
			m.velX += MARIO_ACCEL_SLOWDOWN;
			if (m.velX > 0) m.velX = 0;
		}

		// collision
		// top left
		inblockUL = l.level[m.posX / 2048 * l.sizeY + m.posY / 2048]; // pos / (256*8) --> 2048
		inblockDL = l.level[m.posX / 2048 * l.sizeY + ((m.posY / 256) + 7) / 8];
		inblockUR = l.level[(((m.posX / 256) + 7) / 8) * l.sizeY + m.posY / 2048];
		inblockDR = l.level[(((m.posX / 256) + 7) / 8) * l.sizeY + ((m.posY / 256) + 7) / 8];

		// if / else to avoid moving mario twice
		if (inblockUL != 0) {												// up left
			m.posX = ((int)((int)(lastposX / 256) / 8)) * 8 * 256;
		} else if (inblockDL != 0) {										// down left
			m.posX = ((int)((int)(lastposX / 256) / 8)) * 8 * 256;
		} else if (inblockUR != 0) {										// up right
			m.posX = ((int)(((int)(lastposX / 256) + 7) / 8)) * 8 * 256;
		} else if (inblockDR != 0) {										// down right
			m.posX = ((int)(((int)(lastposX / 256) + 7) / 8)) * 8 * 256;
		}

		// Y
		if (m.velY > 0 && m.velY > MARIO_SPEED_FALL_MAX) m.velY = MARIO_SPEED_FALL_MAX;
		if (m.velY < 0 && m.velY < -MARIO_SPEED_RISE_MAX) m.velY = -MARIO_SPEED_RISE_MAX;
		m.posY += m.velY;

		inblockUL = l.level[m.posX / 2048 * l.sizeY + m.posY / 2048]; // pos / (256*8) --> 2048
		inblockDL = l.level[m.posX / 2048 * l.sizeY + ((m.posY / 256) + 7) / 8];
		inblockUR = l.level[(((m.posX / 256) + 7) / 8) * l.sizeY + m.posY / 2048];
		inblockDR = l.level[(((m.posX / 256) + 7) / 8) * l.sizeY + ((m.posY / 256) + 7) / 8];

		// to make jump more reactive: checks for the block under the feet; gravity is too slow to let mario fall through each frame
		// and Mario can only jump if grounded == 1
		if (l.level[((m.posX / 256) / 8) * l.sizeY + (m.posY / 256) / 8 + 1] != 0 ||
			l.level[(((m.posX / 256) + 7) / 8) * l.sizeY + (m.posY / 256) / 8 + 1] != 0 )
		{
			m.grounded = 1;
		}

		if (inblockUL != 0) {												// up left
			m.posY = ((int)((int)(lastposY / 256) / 8)) * 8 * 256;
			m.bonked = 1;
		} else if (inblockDL != 0) {										// down left
			m.posY = ((int)((int)((lastposY / 256) + 7) / 8)) * 8 * 256;
			m.grounded = 1;
		} else if (inblockUR != 0) {										// up right
			m.posY = ((int)((int)(lastposY / 256) / 8)) * 8 * 256;
			m.bonked = 1;
		} else if (inblockDR != 0) {										// down right
			m.posY = ((int)(((int)(lastposY / 256) + 7) / 8)) * 8 * 256;
			m.grounded = 1;
		}

		if (m.grounded && m.velY > 0) {
			m.velY = 0;
			m.jumping = 0;
		}

		isPressingJump = KeyDown(KEY_CHAR_1);

		if (m.grounded && isPressingJump && !wasHoldingJumpButton) { // jump
			m.velY = -MARIO_ACCEL_JUMP;
			m.jumping = 1;
		} else if (!m.grounded && m.jumping && isPressingJump) {
			m.velY = -MARIO_ACCEL_JUMP + MARIO_ACCEL_JUMP_HOLDA * jumpFrameCount;
		} else if ( (m.jumping && !isPressingJump) || (m.jumping && m.bonked) ) {
			m.jumping = 0;
		}

		if (m.bonked) {
			m.velY = 0;
			if (m.jumping) {
				m.jumping = 0;
			}
		}

		// gravity
		if (!m.grounded && !m.jumping) {
			m.velY += MARIO_ACCEL_GRAVITY;
		}

		if (m.grounded) {
			if (abs(m.velX) < MARIO_SPEED_JUMP_SWITCH) {
				currentJumpLength = MARIO_JUMP_FRAMES_LOW;
			} else {
				currentJumpLength = MARIO_JUMP_FRAMES_HIGH;
			}
		}

		if (m.jumping) {
			++jumpFrameCount;
		} else {
			jumpFrameCount = 0;
		}

		if (jumpFrameCount > currentJumpLength) {
			jumpFrameCount = 0;
			m.jumping = 0;
		}

		lastposX = m.posX;
		lastposY = m.posY;
		m.grounded = 0;
		m.bonked = 0;
		wasHoldingJumpButton = isPressingJump;

		// MARIO END --------------------------------------------------------------------------------------------------------------------------------------
		// ENTITIES START


		// ENTITIES END
		// CAM START
		camX = m.posX/256 - 7 * 8;
		camY = m.posY/256 - 3 * 8;

		if (camX < 0) camX = 0;
		if (camY < 0) camY = 0;
		if (camX > (l.sizeX - 16) * 8) camX = (l.sizeX - 16) * 8;
		if (camY > (l.sizeY - 8) * 8) camY = (l.sizeY - 8) * 8;

		drawScreen(&l, camX / 8, camY / 8, -(camX % 8), -(camY % 8));
		//drawEntities()?
		drawSprite(spritedata + m.spr, m.posX / 256 - camX, m.posY / 256 - camY); // draw Mario
		/*
		 * drawSprite(spritedata + inblockUL * 8, 0, 0);
		 * drawSprite(spritedata + inblockUR * 8, 8, 0);
		 * drawSprite(spritedata + inblockDL * 8, 0, 8);
		 * drawSprite(spritedata + inblockDR * 8, 8, 8);
		 */
		Bdisp_PutDisp_DD();

		// CAM END -----------------------------------------------------------------------------------------------------------------------------------------

		while(lastframetime + 128/30 /* 1sec = 128ticks -> X fps = 128/X */ > RTC_GetTicks()) {
			Sleep(1);
			Bdisp_SetPoint_DD(framesCount, 0, 1);
			++framesCount;
		}

		framesCount = 0;
		lastframetime = RTC_GetTicks();
	}
}

