#pragma once

typedef struct tag_Entity Entity;
typedef struct tag_Level Level;
typedef struct tag_Metadata Metadata;
typedef struct  tag_Mario Mario;

#include <fxlib.h>
#include "sprite.h"
#include "newkeybios.h"
#include <filebios.h>
#include "mysyscall.h"

struct tag_Entity {
	int posX;
	int posY;
	int velX;
	int velY;
	char type;
	char used; // 0 --> unused, 1 --> already spawned. May or may not be used depending on entity
};

struct tag_Metadata {
	int posX;
	int posY;
	Entity entity;
};

struct tag_Level {
	int sizeX;
	int sizeY;
	int sizeEntities;
	int sizeMetadata;
	char* level; // pointer to sprites: each is an index pointing inside the sprite array
	Entity* entities; // each Entity's struct is copied at runtime
	Metadata* metadata;

};

struct tag_Mario {
	int posX; // mario position throughout, in pixel world coords.
	int posY; // also used for camera positioning.
	int velX; // /!\ accel values are 256 for 1 block
	int velY; // positive is down, neg is up
	int spr; // current mario sprite, used for animation
	char grounded; // 1 if on ground last frame, else 0
	char bonked; // 1 if hit head on object above it
	char jumping;
	char running;
};

static const FONTCHARACTER defaultPath[] = {'\\','\\','f','l','s','0','\\','m','m','k','.','d','a','t',0};

int checkErrors(const int err);
int saveLevel(Level l, const FONTCHARACTER* path);
Level loadLevel(const FONTCHARACTER* path);
Level newLevel(int sizeX, int sizeY, int sizeEntities, int sizeMetadata);
void freeLevel(Level* l);
void editLevel();
void playLevel(Level l, int x, int y);

// /!\ accel values are 256 for 1 block

#define MARIO_SPEED_WALK_MIN (1 * 256 / 4) // unused
#define MARIO_SPEED_WALK_MAX (1 * 256)
#define MARIO_SPEED_RUN_MAX (2 * 256) // unsure
#define MARIO_ACCEL_WALK (1 * 256 / 4)
#define MARIO_ACCEL_SLOWDOWN (3 * 256 / 16)
#define MARIO_ACCEL_RUN	 (1 * 256 / 4)// TODO: run
#define MARIO_ACCEL_JUMP (9 * 256 / 4)
#define MARIO_ACCEL_JUMP_HOLDA (3 * 256 / 64)
#define MARIO_JUMP_FRAMES_HIGH (20)
#define MARIO_JUMP_FRAMES_LOW (15)
#define MARIO_SPEED_JUMP_SWITCH (1 * 256) 
#define MARIO_ACCEL_GRAVITY (1 * 256 / 4)
#define MARIO_SPEED_FALL_MAX (3 * 256)
#define MARIO_SPEED_RISE_MAX (4 * 256)
